/* 
   Copyright (c) 1995 by Cisco systems, Inc.
   All rights reserved.

   Please NOTE:  None of the TACACS code available here comes with any
   warranty or support.
*/

#include "tac_plus.h"
#include "expire.h"

/* internal state variables */
#define STATE_AUTHEN_START   0	/* no requests issued */
#define STATE_AUTHEN_GETUSER 1	/* username has been requested */
#define STATE_AUTHEN_GETPASS 2	/* password has been requested */

static int pw_verify();

/* Set data->status depending on whether a user has expired or not */

static void
set_expiration_status(exp_date, data)
char *exp_date;
struct authen_data *data;
{
    /* Check the expiration date, if any. If NULL, this check will return
     * PW_OK */
    int expired = check_expiration(exp_date);

    switch (expired) {
    case PW_OK:
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "Password has not expired %s", 
		   exp_date ? exp_date : "<no expiry date set>");

	data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
	return;

    case PW_EXPIRING:
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "Password will expire soon %s", 
		   exp_date ? exp_date : "<no expiry date set>");
	if (data->server_msg)
	    free(data->server_msg);
	data->server_msg = tac_strdup("Password will expire soon");
	data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
	return;

    case PW_EXPIRED:
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "Password has expired %s", 
		   exp_date ? exp_date : "<no expiry date set>");
	if (data->server_msg)
	    free(data->server_msg);
	data->server_msg = tac_strdup("Password has expired");
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return;

    default:
	report(LOG_ERR, "%s: Bogus return value %d from check_expiration", 
	       session.peer, expired);
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	return;
    }
}

#ifdef SKEY
#include <skey.h>
#endif

struct private_data {
#ifdef SKEY
    struct skey skey;
#endif
    char password[MAX_PASSWD_LEN + 1];
    int state;
};

#ifdef SKEY

/* Use s/key to verify a supplied password using state set up earlier
when the username was supplied */

static int
skey_verify(passwd, data)
char *passwd;
struct authen_data *data;
{
    struct private_data *p = data->method_data;
    struct skey *skeyp = &p->skey;

    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

    if (skeyverify(skeyp, passwd) == 0) {
	/* S/Key authentication succeeded */
	data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
	if (skeyp->n < 5) {
	    data->server_msg = tac_strdup("Password will expire soon");
	    return (1);
	}
    }
    return (0);
}

#endif				/* SKEY */

static int passwd_verify();

/* Verify that this user/password is valid. Return 1 if valid */

static int
verify(name, passwd, data, recurse)
char *name, *passwd;
struct authen_data *data;
int recurse;
{
    char *exp_date;
    char *cfg_passwd = cfg_get_login_secret(name, recurse);
    char *p;

    /* If there is no login password for this user, see if there is 
       a global password for her that we can use */
    if (!cfg_passwd) {
	cfg_passwd = cfg_get_global_secret(name, recurse);
    }

    /* If there is still no password for this user (or no user for that
       matter) make a last ditch attempt is to use a password file set
       by the default authentication = file <file> statement */

    if (!cfg_passwd) {
	char *file = cfg_get_authen_default();
	if (file) {
	    return (passwd_verify(name, passwd, data, file));
	}
	/* otherwise, fail */
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return (0);
    }

    p = tac_find_substring("cleartext ", cfg_passwd);
    if (p) {
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "verify daemon %s == NAS %s", p, passwd);

	/* Only first eight characters are significant */
	if (strncmp(passwd, p, 8)) {
	    if (debug & DEBUG_PASSWD_FLAG)
	       report(LOG_DEBUG, "verify returns 0"); 
	    return(0);
	}

	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "verify returns 1"); 

	exp_date = cfg_get_expires(name, recurse);
	set_expiration_status(exp_date, data);
	return (data->status == TAC_PLUS_AUTHEN_STATUS_PASS);
    }

    p = tac_find_substring("file ", cfg_passwd);
    if (p) {
	return (passwd_verify(name, passwd, data, p));
    }

    if (STREQ(cfg_passwd, "skey")) {

#ifdef SKEY
	return (skey_verify(passwd, data));
#else				/* SKEY */
	report(LOG_ERR, 
	       "%s: Error Attempt to authenticate skey user %s using non-skey version",
	       session.peer, name);
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return (0);
#endif				/* SKEY */
    }

    p = tac_find_substring("des ", cfg_passwd);
    if (p) {
	/* try to verify the password */
	if (!pw_verify(passwd, p))
	    return (0);

	exp_date = cfg_get_expires(name, recurse);
	set_expiration_status(exp_date, data);
	return (data->status == TAC_PLUS_AUTHEN_STATUS_PASS);
    }

    /* Oops. No idea what kind of password this is. This should never
       happen. */

    report(LOG_ERR, "%s: Error cannot identify password type %s for %s",
	   session.peer, 
	   cfg_passwd ? cfg_passwd : "<NULL>", 
	   name ? name : "<unknown>");

    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
    return (0);
}

/* verify that this user/password is valid per a passwd(5) style
   database. Return 0 if invalid. */

static int
passwd_verify(user, supplied_passwd, data, filename)
char *user, *supplied_passwd;
struct authen_data *data;
char *filename;
{
    struct passwd *pw;
    char *exp_date;

    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

    setpwent();

    if (filename && !STREQ(filename, "/etc/passwd")) {
	/* an alternate filename */
	if (!(access(filename, R_OK) == 0)) {
	    report(LOG_ERR, "%s: %s Cannot access %s %s",
		   session.peer, user, 
		   filename, sys_errlist[errno]);
	    return (0);
	}
	endpwent();
	if (setpwfile(filename) != 1) {
	    report(LOG_ERR, "%s: Cannot setpwfile %s %s",
		   session.peer, filename, sys_errlist[errno]);
	    return (0);
	}
    }
    pw = getpwnam(user);
    endpwent();

    if (pw == NULL)
	/* no entry exists */
	return (0);

    if (*pw->pw_passwd == NULL ||
	supplied_passwd == NULL ||
	*supplied_passwd == NULL) {
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return (0);
    }
    /* try to verify the password */
    if (!pw_verify(supplied_passwd, pw->pw_passwd)) {
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return (0);
    }
    /* password ok. Check expiry field */
    exp_date = pw->pw_shell;

    set_expiration_status(exp_date, data);
    return (data->status == TAC_PLUS_AUTHEN_STATUS_PASS);
}

/* verify that this user/password may login validating her
   password. Recurse up groups if needed.

   Sets data->status */

static void
login(name, passwd, data)
char *name, *passwd;
struct authen_data *data;
{
    (void) verify(name, passwd, data, TAC_PLUS_RECURSE);
}

/* Verify that this user/password may ENABLE on this NAS as follows:

   Pass or fail is indicated by setting data->status

   Note: usernames are not used when enabling.
*/

static void
enable(passwd, data)
char *passwd;
struct authen_data *data;
{
    int level = data->NAS_id->priv_lvl;

    /* sanity check */
    if (level < TAC_PLUS_PRIV_LVL_MIN || level > TAC_PLUS_PRIV_LVL_MAX) {
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	data->server_msg = tac_strdup("Invalid privilege level in packet");
	report(LOG_ERR, "%s level=%d %s", session.peer, level, data->server_msg);
	return;
    }
    /* 0 <= level <= 14: look for $enab<n>$ and verify */
    if (level < TAC_PLUS_PRIV_LVL_MAX) {
	char buf[11];

	sprintf(buf, "$enab%d$", level);
	if (!verify(buf, passwd, data, TAC_PLUS_NORECURSE))
	    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return;
    }

    /* 2). level=15. Try $enab15$ or $enable$ (for backwards
       compatibility) and verify */

    if (verify("$enable$", passwd, data, TAC_PLUS_NORECURSE) ||
	verify("$enab15$", passwd, data, TAC_PLUS_NORECURSE)) {
	return;
    }

    /* return fail */
    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

    return;
}

/*
 * verify a provided password against an encrypted one
 * return 1 if verified, 0 otherwise
 */

static int
pw_verify(users_passwd, encrypted_passwd)
char *users_passwd, *encrypted_passwd;
{
    char *ep;

    if (debug & DEBUG_PASSWD_FLAG)
	report(LOG_DEBUG, "verify %s %s", users_passwd, encrypted_passwd);

    if (users_passwd == NULL ||
	*users_passwd == NULL ||
	encrypted_passwd == NULL ||
	*encrypted_passwd == NULL) {
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "verify returns 0");
	return (0);
    }
    ep = (char *) crypt(users_passwd, encrypted_passwd);

    if (debug & DEBUG_PASSWD_FLAG)
	report(LOG_DEBUG, "%s encrypts to %s", users_passwd, ep);

    if (strcmp(ep, encrypted_passwd) == 0) {
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "verify returns 1");
	return (1);
    }
    if (debug & DEBUG_PASSWD_FLAG)
	report(LOG_DEBUG, "verify returns 0");

    return (0);
}

/*
 * Generic tacacs login authentication function. Wants a username
 * and a password, and tries to verify them.
 *
 * Choose_authen will ensure that we already have a username before this
 * gets called.
 *
 * We will query for a password and keep it in the method_data.
 *
 * Any strings returned via pointers in authen_data must come from the
 * heap. They will get freed by the caller.
 *
 * Return 0 if data->status is valid, otherwise 1
 */

int
authenticate_fn(data)
struct authen_data *data;
{
    char *name, *passwd;
    struct private_data *p;
    char *prompt;

    p = (struct private_data *) data->method_data;

    /* An abort has been received. Clean up and return */
    if (data->flags & TAC_PLUS_CONTINUE_FLAG_ABORT) {
	if (data->method_data)
	    free(data->method_data);
	data->method_data = NULL;
	return (1);
    }
    /* Initialise method_data if first time through */
    if (!p) {
	p = (struct private_data *) tac_malloc(sizeof(struct private_data));
	bzero(p, sizeof(struct private_data));
	data->method_data = p;
	p->state = STATE_AUTHEN_START;
    }

    /* Unless we're enabling, we need a username */
    if (data->service != TAC_PLUS_AUTHEN_SVC_ENABLE &&
	!(char) data->NAS_id->username[0]) {
	switch (p->state) {

	case STATE_AUTHEN_GETUSER:
	    /* we have previously asked for a username but none came back.
	     * This is a gross error */
	    data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	    report(LOG_ERR, "%s: No username supplied after GETUSER",
		   session.peer);
	    return (0);

	case STATE_AUTHEN_START:
	    /* No username. Try requesting one */
	    data->status = TAC_PLUS_AUTHEN_STATUS_GETUSER;
	    if (data->service == TAC_PLUS_AUTHEN_SVC_LOGIN) {
		prompt = "\nUser Access Verification\n\nUsername: ";
	    } else {
		prompt = "Username: ";
	    }
	    data->server_msg = tac_strdup(prompt);
	    p->state = STATE_AUTHEN_GETUSER;
	    return (0);

	default:
	    /* something awful has happened. Give up and die */
	    report(LOG_ERR, "%s: authenticate_fn bad state %d", 
		   session.peer, p->state);
	    return (1);
	}
    }

    /* we now have a username if we needed one */
    name = data->NAS_id->username;

    /* Do we have a password? */
    passwd = p->password;

    if (!passwd[0]) {

	/* no password yet. Either we need to ask for one and expect to get
	 * called again, or we asked but nothing came back, which is fatal */

	switch (p->state) {
	case STATE_AUTHEN_GETPASS:
	    /* We have already asked for a password. This should be the reply */
	    strncpy(passwd, data->client_msg, MAX_PASSWD_LEN);
	    passwd[MAX_PASSWD_LEN + 1] = '\0';
	    break;

	default:
	    /* Request a password */
	    prompt = NULL;

#ifdef SKEY
	    /* Does this user use skey? */
	    passwd = cfg_get_login_secret(name, TAC_PLUS_RECURSE);
	    if (passwd && STREQ(passwd, "skey")) {
		char skeyprompt[80];
		if (skeychallenge(&p->skey, name, skeyprompt) == 0) {
		    char buf[256];
		    sprintf(buf, "%s\nPassword: ", skeyprompt);
		    prompt = tac_strdup(buf);
		}
	    }
#endif /* SKEY */

	    if (!prompt) {
		data->flags = TAC_PLUS_AUTHEN_FLAG_NOECHO;
		prompt = tac_strdup("Password: ");
	    }

	    data->server_msg = prompt;
	    data->status = TAC_PLUS_AUTHEN_STATUS_GETPASS;
	    p->state = STATE_AUTHEN_GETPASS;
	    return (0);
	}
    }
    /* We have a username and password. Try validating */

    /* Assume the worst */
    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

    switch (data->service) {
    case TAC_PLUS_AUTHEN_SVC_LOGIN:
    case TAC_PLUS_AUTHEN_SVC_PPP:
	login(name, passwd, data);
	if (debug)
	    report(LOG_INFO, "login query for '%s' %s from %s %s",
		   name && name[0] ? name : "unknown",
		   data->NAS_id->NAS_port && data->NAS_id->NAS_port[0] ?
		       data->NAS_id->NAS_port : "unknown",
		   session.peer,
		   (data->status == TAC_PLUS_AUTHEN_STATUS_PASS) ?
		   "accepted" : "rejected");
	break;

    case TAC_PLUS_AUTHEN_SVC_ENABLE:
	enable(passwd, data);
	if (debug)
	    report(LOG_INFO, "enable query for '%s' %s from %s %s",
		   name && name[0] ? name : "unknown",
		   data->NAS_id->NAS_port && data->NAS_id->NAS_port[0] ?
		       data->NAS_id->NAS_port : "unknown",
		   session.peer, 
		   (data->status == TAC_PLUS_AUTHEN_STATUS_PASS) ?
		   "accepted" : "rejected");
	break;

    default:
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	report(LOG_ERR, "%s: Bogus service value %d from packet", 
	       session.peer, data->service);
	break;
    }

    if (data->method_data)
	free(data->method_data);
    data->method_data = NULL;

    switch (data->status) {
    case TAC_PLUS_AUTHEN_STATUS_ERROR:
    case TAC_PLUS_AUTHEN_STATUS_FAIL:
    case TAC_PLUS_AUTHEN_STATUS_PASS:
	return (0);
    default:
	report(LOG_ERR, "%s: authenticate_fn couldn't set recognizable status %d",
	       session.peer, data->status);
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	return (1);
    }
}
